REBOL [
	author: 'chl
	dates: [2003-01-02]
	description: "very simple main-memory based triple-space implementation"
]

space: context [

	export: func [words] [
		foreach word words [
			set in system/words word get word
		]
	]

	type: 		1
	subject: 	2
	predicate: 	3
	object: 	4
	annotations: 	5

	export [
		type subject predicate object annotations
	]

	file: none
	triples: copy []
	snapshot-behaviour: none ; none or 'always


	load: func [file snapshot-behaviour] [
		self/file: file
		if none? file [
			self/triples: copy []
			return false
		]
		if not exists? file [
			save file copy []
		]
		self/snapshot-behaviour: snapshot-behaviour
		self/triples: system/words/load file
	]

	; layer sub-zero

	to-lit: func [x] [
		if word? x [return to-lit-word x] 
		if path? x [return to-lit-path x]
		x
	]

	replace-deep: func [target search replace /lit] [
		forskip target 1 [
			either (block? target/1) or (paren? target/1) [
				change/only target replace-deep target/1 search replace 
			] [
				if target/1 = search [
					either lit [
						change/only target to-lit replace
					] [
						change/only target replace
					]
				]
			]
		]
		target: head target
	]

	copy-series: func [x] [
		either series? x [
			return copy/deep x
		] [
			return copy reduce [false]
		]
	]

	unique-triples: func [triples /local attributes result] [
		result: copy []
		attributes: copy []
		foreach triple unique triples [
			either triple/:type = 'attribute [
				either not found? find/only attributes reduce [triple/:subject triple/:predicate] [
					repend/only attributes [triple/:subject triple/:predicate]
					append/only result triple
				] [
					self/remove triple
				]
			] [
				append/only result triple
			]
		]
		result
	]

	none?: func [x] [
		(system/words/none? x) or ('none = x)
	]

	; basic functions

	valid?: func [triple] [
		either any [
			(length? triple) < 4
			not found? find [attribute link] triple/:type
			not string? triple/:subject
		] [
			false
		] [
			true
		]
	]

	snapshot: does [
		if not none? self/file [
			save self/file self/triples
		]
	]

	snapshot-if-necessary: does [
		if 'always = self/snapshot-behaviour [
			self/snapshot
		]
	]

	add-unchecked: func [triple] [
		insert/only self/triples triple
		triple
	]

	clear: does [
		system/words/clear self/triples
		self/snapshot-if-necessary
		
	]

	add: func [triple] [
		if not valid? triple [
			return false
		]
		if found? find/only self/triples triple [ ; @@ necessary?
			return false
		]
		self/triples: unique-triples head insert/only self/triples triple
		self/snapshot-if-necessary
		triple
	]

	remove: func [triple] [
		system/words/remove find/only self/triples triple
		self/snapshot-if-necessary
		triple
	]

	match: func [template f /with-indices /local result] [
		result: copy []
		foreach triple self/triples [
			if all [ ; @@ scrap infix once /view is bug-fixed
				any [(self/none? template/:type) (template/:type == triple/:type)]
			    	any [(self/none? template/:subject) (template/:subject == triple/:subject)]
			    	any [(self/none? template/:predicate) (template/:predicate == triple/:predicate)]
			    	any [(self/none? template/:object) (template/:object == triple/:object)] 
			] [
				append/only result triple
			]
		]
		foreach r result [
			f r
		]
		none
	]

	; space/query [none [(x = "Vanilla") or (x = "chris")] none none]

	query: func [template f /local result] [
		result: copy []
		foreach triple triples [
			if all [
			 	any [(self/none? template/:type) (do self/replace-deep/lit self/copy-series template/:type 'x triple/:type)]
			    	any [(self/none? template/:subject) (do self/replace-deep/lit self/copy-series template/:subject 'x triple/:subject)]
				any [(self/none? template/:predicate) (do self/replace-deep/lit self/copy-series template/:predicate 'x triple/:predicate)]
				any [(self/none? template/:object) (do self/replace-deep/lit self/copy-series template/:object 'x triple/:object)]
			] [
				append/only result triple
			]
		]
		foreach r result [
			f r
		]
		none
	]

	; shortcuts & sugar

	exists-subject?: func [subject] [
		found? find self/subjects subject
	]

	subjects: func [/local subjects] [
		subjects: copy []
		foreach triple self/triples [
			append subjects triple/:subject
		]
		unique subjects
	]

	build: func [method template f /local triples] [
		triples: copy []
		self/:method template func [triple] [
			append/only triples f triple
		]
		triples
	]

	fetch: func [method template] [
		self/build method template func [x] [x]
	]

	probe: func [method template] [
		self/:method template func [x] [system/words/probe x]
	]

	for-all: func [triples method /local fomer-behaviour] [
		former-behaviour: self/snapshot-behaviour
		self/snapshot-behaviour: none
		foreach triple self/triples [
			self/:method triple
		]
		self/snapshot-behaviour: former-behaviour
		self/snapshot-if-necessary
		self/triples
	]

	add-all: func [triples /local fomer-behaviour] [
		append self/triples triples
		self/triples: unique-triples self/triples
		self/snapshot-if-necessary
	]

	remove-all: func [triples /local fomer-behaviour] [
		self/for-all triples 'remove
	]

	get-object: func [subject predicate] [
		foreach triple self/triples [
			if all [
				triple/:subject = subject
				triple/:predicate = predicate
			] [
				return triple/:object
			]
		]
	]

]

if found? find to-string system/options/script %3space-in-memory-lib.r [
	halt
]

